/*
 * livepatch_bsc1190127
 *
 * Fix for CVE-2021-31440, bsc#1190127
 *
 *  Upstream commit:
 *  10bf4e83167c ("bpf: Fix propagation of 32 bit unsigned bounds from 64 bit
 *                 bounds")
 *
 *  SLE12-SP3 commit:
 *  not affected
 *
 *  SLE12-SP4, SLE12-SP5, SLE15 and SLE15-SP1 commit:
 *  not affected
 *
 *  SLE15-SP2 commit:
 *  not affected
 *
 *  SLE15-SP3 commit:
 *  26f1fe9cc8fcbc766522eff945a82f90ff74c450
 *
 *
 *  Copyright (c) 2021 SUSE
 *  Author: Nicolai Stange <nstange@suse.de>
 *
 *  Based on the original Linux kernel code. Other copyrights apply.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

/* klp-ccp: from kernel/bpf/verifier.c */
#include <uapi/linux/btf.h>
#include <linux/kernel.h>
#include <linux/types.h>
#include <linux/slab.h>
#include <linux/bpf.h>
#include <linux/bpf_verifier.h>
#include <linux/filter.h>
#include <net/netlink.h>
#include <linux/file.h>
#include <linux/vmalloc.h>
#include <linux/stringify.h>
#include <linux/error-injection.h>
/* klp-ccp: from kernel/bpf/disasm.h */
#include <linux/bpf.h>
#include <linux/kernel.h>
#include <linux/stringify.h>

/* klp-ccp: from kernel/bpf/verifier.c */
static void __mark_reg32_unbounded(struct bpf_reg_state *reg)
{
	reg->s32_min_value = S32_MIN;
	reg->s32_max_value = S32_MAX;
	reg->u32_min_value = 0;
	reg->u32_max_value = U32_MAX;
}

static void (*klpe___update_reg_bounds)(struct bpf_reg_state *reg);

static void (*klpe___reg32_deduce_bounds)(struct bpf_reg_state *reg);

static void (*klpe___reg64_deduce_bounds)(struct bpf_reg_state *reg);

static void klpr___reg_deduce_bounds(struct bpf_reg_state *reg)
{
	(*klpe___reg32_deduce_bounds)(reg);
	(*klpe___reg64_deduce_bounds)(reg);
}

static void (*klpe___reg_bound_offset)(struct bpf_reg_state *reg);

static bool __reg64_bound_s32(s64 a)
{
	return a > S32_MIN && a < S32_MAX;
}

static bool klpp___reg64_bound_u32(u64 a)
{
	/*
	 * Fix CVE-2021-31440
	 *  -3 lines, +1 line
	 */
	return a > U32_MIN && a < U32_MAX;
}

void klpp___reg_combine_64_into_32(struct bpf_reg_state *reg)
{
	__mark_reg32_unbounded(reg);

	if (__reg64_bound_s32(reg->smin_value) && __reg64_bound_s32(reg->smax_value)) {
		reg->s32_min_value = (s32)reg->smin_value;
		reg->s32_max_value = (s32)reg->smax_value;
	}
	/*
	 * Fix CVE-2021-31440
	 *  -4 lines, +4 lines
	 */
	if (klpp___reg64_bound_u32(reg->umin_value) && klpp___reg64_bound_u32(reg->umax_value)) {
		reg->u32_min_value = (u32)reg->umin_value;
		reg->u32_max_value = (u32)reg->umax_value;
	}

	/* Intersecting with the old var_off might have improved our bounds
	 * slightly.  e.g. if umax was 0x7f...f and var_off was (0; 0xf...fc),
	 * then new var_off is (0; 0x7f...fc) which improves our umax.
	 */
	klpr___reg_deduce_bounds(reg);
	(*klpe___reg_bound_offset)(reg);
	(*klpe___update_reg_bounds)(reg);
}



#include <linux/kernel.h>
#include <linux/module.h>
#include "livepatch_bsc1190127.h"
#include "../kallsyms_relocs.h"

static struct klp_kallsyms_reloc klp_funcs[] = {
	{ "__update_reg_bounds", (void *)&klpe___update_reg_bounds },
	{ "__reg32_deduce_bounds", (void *)&klpe___reg32_deduce_bounds },
	{ "__reg64_deduce_bounds", (void *)&klpe___reg64_deduce_bounds },
	{ "__reg_bound_offset", (void *)&klpe___reg_bound_offset },
};

int livepatch_bsc1190127_init(void)
{
	return __klp_resolve_kallsyms_relocs(klp_funcs, ARRAY_SIZE(klp_funcs));
}
